

#pragma ident "@(#)connect.c	1.6 95/05/24 SMI"

/*
 * The user presses the mouse on the drawing area and moves it.  When the user
 * lets go of the mouse button, the intermediate points get connected in the
 * current color.
 *
 * This is an embellishment of a program in Charles Petzold's "Programming
 * Windows 3.1".
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <DPS/dpsXclient.h>
#include "wraps.h"


#define NUM_COLORS	18
#define MAX_POINTS	100

/* XID of default GC from a given widget */
#define GCXID(w)	(XGContextFromGC(DefaultGCOfScreen(XtScreen(w))))


/* like XPoint, but for PS coordinates */
typedef struct {
    float x, y;
} UserPoint;


typedef struct {
    Widget 	toplevel;
    Widget	form;
    Widget	clear_btn;
    Widget	draw_area;
    Widget	rc;
    Widget	color_btns[NUM_COLORS];	/* color buttons */
    Pixmap	pixmaps[NUM_COLORS];	/* swatches for color buttons */
    XColor	colors[NUM_COLORS];	/* color names */
    UserPoint	points[MAX_POINTS];	/* points plotted */
    int		num_points;		/* count of points plotted */
    float	invctm[6];		/* inverse transform mtx */
    int		y_offset;		/* y offset for coordinate conv. */
    DPSContext	ctxt;
} UI;



/* convert X coordinate to PostScript User Space */
void
convert_x_to_ps(
	UI* ui,
	int x, 
	int y,
	float* ux,
	float* uy
)
{
    float iy;

    iy = y - ui->y_offset;
    *ux = ui->invctm[0] * x + ui->invctm[2] * iy + ui->invctm[4];
    *uy = ui->invctm[1] * x + ui->invctm[3] * iy + ui->invctm[5];
}


static DPSContext
dps_context_from_widget(
	Widget w
)
{
    Dimension height;
    DPSContext ctxt;

    XtVaGetValues(w, XtNheight, &height, NULL);

    ctxt 
	= XDPSCreateSimpleContext(XtDisplay(w), XtWindow(w),
				  DefaultGCOfScreen(XtScreen(w)),
				  0, height,
				  DPSDefaultTextBackstop,	/* send output to stdout */
				  DPSDefaultErrorProc,		/* send errors to stdout */
				  (DPSSpace) NULL		/* create a new space */
		  );

    /* libdps requires this when used with Xt */
    XDPSSetEventDelivery(XtDisplay(w), dps_event_pass_through);

    return ctxt;
}



/* activateCallback on color buttons */
static void
color_picked(
	Widget w,
	UI* ui,
	XmPushButtonCallbackStruct* cbs
)
{
    int ii;

    for(ii=0; ui->color_btns[ii] != w; ++ii) {
	/* do nothing, just looking for the right widget */;
    }

    DPSsetrgbcolor(ui->ctxt, 
		   /* divide by 256 to scale the numbers between 0-1 */
		   (float)ui->colors[ii].red   / 256.0,
		   (float)ui->colors[ii].green / 256.0,
		   (float)ui->colors[ii].blue  / 256.0
	);
}


static void
connect_points(
	UI* ui 
)
{
    DPSContext ctxt = ui->ctxt;
    int ystart = 1;
    int xx, yy;

    if ( ui->num_points == 0 ) {
	return;
    }

    /* connect each point to each other point, trying to
     * avoid duplicate lines for better performance.
     */
    for(xx=0; xx < ui->num_points; ++xx) {
	for (yy=ystart; yy < ui->num_points; ++yy) {
	    PSWDrawLine(ctxt,
			ui->points[xx].x, ui->points[xx].y,
			ui->points[yy].x, ui->points[yy].y
		);
	}
	ystart++;
    }

    DPSstroke(ctxt);
    DPSFlushContext(ctxt);
}



/* XmNexposeCallback on "drawing_a" DrawingArea */
static void
exposed(
	Widget w, 
	UI* ui,
	XmDrawingAreaCallbackStruct* cbs
)
{
    connect_points( ui );
}


/* XmNactivateCallback on "Clear" PushButton */
static void
clear_points(
	Widget w,
	UI* ui,
	XmPushButtonCallbackStruct* cbs
)
{
    ui->num_points = 0;
    XClearWindow(XtDisplay(ui->draw_area), XtWindow(ui->draw_area));
}


/* action proc on "drawing_a" DrawingArea */
static void
collect(
	Widget w,
	XMotionEvent* event,
	String *args, 
	int* argcount
)
{
    float ux, uy;
    UI *ui;

    XtVaGetValues(w, XmNuserData, &ui, NULL);


    /* up event -- connect points */
    if ( strcmp(args[0], "up") == 0 ) {
	connect_points( ui );
	return;
    }


    /* down event -- clear list of points */
    if ( strcmp(args[0], "down") == 0 ) {
	clear_points(w, ui, NULL);
    }

    if ( ui->num_points >= MAX_POINTS ) {
	return;
    }

    /* Collect points from motion events */
    convert_x_to_ps(ui, event->x, event->y, &ux, &uy);
    ui->points[ui->num_points].x = ux;
    ui->points[ui->num_points].y = uy;
    PSWDrawPoint(ui->ctxt,
		ui->points[ui->num_points].x, 
		ui->points[ui->num_points].y
	);
    DPSstroke(ui->ctxt);
    DPSFlushContext(ui->ctxt);
    ui->num_points++;
}



int
main(
	int argc, 
	char** argv
)
{
    int ii;
    UI ui = {0};
    XtAppContext app;
    XtActionsRec actions;
    Dimension height;
    const String fallback_resources[] = {
	"connect.title: Connect-The-Dots",
	"*.clear_btn.labelString: Clear",
	"*.drawing_a.background: white",
	"*.drawing_a.height: 400",
	"*.drawing_a.borderWidth: 1",
	"*.drawing_a.topOffset: 3",
	"*.drawing_a.leftOffset: 3",
	"*.drawing_a.rightOffset: 3",
	"*.drawing_a.bottomOffset: 3",
	NULL
    };
    String translations =
        "<Btn1Down>:   collect(down)\n\
         <Btn1Up>:     collect(up)  \n\
         <Btn1Motion>: collect(motion)";
    const char* const color_names[] = {
	"Black", "Red", "Green", "Blue", "SlateGray", "Navy", "Orange",
	"Yellow", "Pink", "Magenta", "Cyan", "Brown", "Grey", "LimeGreen",
	"Turquoise", "Violet", "Wheat", "Purple"
    };


    ui.toplevel 
	= XtVaAppInitialize(&app, "Connect", NULL, 0, 
			    &argc, argv, 
			    (String*)fallback_resources, 
			    NULL);

    ui.form
	= XtVaCreateManagedWidget("form", xmFormWidgetClass, ui.toplevel,
				  NULL);


    ui.clear_btn
	= XtVaCreateManagedWidget("clear_btn", xmPushButtonWidgetClass, ui.form,
				  XmNtopAttachment,	XmATTACH_FORM,
				  XmNleftAttachment,	XmATTACH_FORM,
				  XmNrightAttachment,	XmATTACH_FORM,
				  NULL);
    XtAddCallback(ui.clear_btn, XmNactivateCallback, (XtCallbackProc)clear_points, &ui);


    /* manager for color buttons */
    ui.rc
	= XtVaCreateManagedWidget("rc", xmRowColumnWidgetClass, ui.form,
				  XmNorientation,	XmHORIZONTAL,
				  XmNpacking,		XmPACK_COLUMN,
				  XmNtopAttachment,	XmATTACH_WIDGET,
				  XmNtopWidget,		ui.clear_btn,
				  XmNleftAttachment,	XmATTACH_FORM,
				  NULL);

    /* create the color swatches and buttons */
    for (ii=0; ii<NUM_COLORS; ++ii ) {
        ui.pixmaps[ii] 
	    = XCreatePixmap(XtDisplay(ui.rc), RootWindowOfScreen(XtScreen(ui.rc)),
			    16, 16, DefaultDepthOfScreen(XtScreen(ui.rc))
			    );
	ui.color_btns[ii]
	    = XtVaCreateManagedWidget(color_names[ii], xmPushButtonWidgetClass, ui.rc,
				      XmNlabelType,	XmPIXMAP,
				      XmNlabelPixmap,	ui.pixmaps[ii], 
				      NULL);

        XtAddCallback(
		ui.color_btns[ii],
		XmNactivateCallback, (XtCallbackProc)color_picked, 
		&ui
	);
    }


    /* XmNinputCallback won't give motion events, so use translations */
    actions.string = "collect";
    actions.proc = (XtActionProc)collect;
    XtAppAddActions(app, &actions, 1);
    ui.draw_area 
	= XtVaCreateManagedWidget("drawing_a",
				  xmDrawingAreaWidgetClass, ui.form,
				  XmNtranslations,
					XtParseTranslationTable(translations),
				  XmNuserData,		&ui,
				  XmNtopAttachment,	XmATTACH_WIDGET,
				  XmNtopWidget,		ui.rc,
				  XmNleftAttachment,	XmATTACH_FORM,
				  XmNrightAttachment,	XmATTACH_FORM,
				  XmNbottomAttachment,	XmATTACH_FORM,
				  NULL);
    XtAddCallback(ui.draw_area, XmNexposeCallback, (XtCallbackProc)exposed, &ui);

    XtRealizeWidget(ui.toplevel);

    /*
     * Do DPS initialization ofter Realize phase so we have window id.
     * In some cases it may make more sense to do this the first time
     * the exposeCallback is executed.
     */
    ui.ctxt = dps_context_from_widget( ui.draw_area );

#ifdef DEBUG
    {
	DPSContext tc
	    = DPSCreateTextContext(
			DPSDefaultTextBackstop, 
			DPSDefaultErrorProc
		);
	DPSChainContext(ui.ctxt, tc);
    }
#endif

    for (ii=0; ii<NUM_COLORS; ++ii) {
	Display *dpy = XtDisplay(ui.draw_area);
	Colormap cmap = DefaultColormapOfScreen(XtScreen(ui.draw_area));

	/* Would be faster to access the X color database directly. */
	if (XParseColor(dpy, cmap, color_names[ii], &ui.colors[ii]) == 0) {
	    char buf[32];
	    sprintf(buf, "Can't find color %s", color_names[ii]);
	    XtWarning(buf);
	    exit(-1);
	}

	/* These are returned in 16 bit values, even
	 * though only 8 bits are significant.
	 */
	ui.colors[ii].red   >>= 8;
	ui.colors[ii].green >>= 8;
	ui.colors[ii].blue  >>= 8;

	DPSsetXgcdrawable(ui.ctxt,
			  /* the wrap needs the XID of the GC, not the _XGC* */
			  GCXID(ui.draw_area),
			  ui.pixmaps[ii], 0, 16
	    );
	DPSsetrgbcolor(ui.ctxt, 
		       /* divide by 256 to scale the numbers between 0-1 */
		       (float)ui.colors[ii].red   / 256.0,
		       (float)ui.colors[ii].green / 256.0,
		       (float)ui.colors[ii].blue  / 256.0
	    );

	DPSrectfill(ui.ctxt, 0.0, 0.0, 16.0, 16.0);
    } /* -- end for loop */

    XtVaGetValues(ui.draw_area, XtNheight, &height, NULL);
    DPSsetXgcdrawable(ui.ctxt,
		      GCXID(ui.draw_area),
		      XtWindow(ui.draw_area), 
		      0, height
	);

    /* Needed for X to PS coordinate conversions. */
    PSWGetInverseCTM(ui.ctxt, ui.invctm);
    ui.y_offset = height;	/* this is true because we just set it, else
				 * we'd have to query the currentXoffset.
				 */

    DPSsetrgbcolor(ui.ctxt,	/* default to first color (black) */
		   (float)ui.colors[0].red   / 256.0,
		   (float)ui.colors[0].green / 256.0,
		   (float)ui.colors[0].blue  / 256.0
	);


    while ( True ) {
	/* DPS friendly equivalent of XtAppMainLoop() */
	XEvent event;
	XtAppNextEvent(app, &event);
	if ( !XDPSDispatchEvent(&event) )
	    (void) XtDispatchEvent(&event);
    }
}
